"""
Interfaces used by the Workspaces' capabilities system.
"""

import inspect
from abc import ABC, abstractmethod
from typing import List, Union, Optional, Any, Type

from .product_interfaces import ProductIF, FutureProductIF

ProductLocator = str
CapabilityName = str


class CapabilityIF:
    id: int
    name: str
    steps: List["CapabilityStepIF"]
    max_jobs: int

    def create_request(
        self, parameters: "ParameterIF", future_products: FutureProductIF
    ) -> "CapabilityRequestIF":
        raise NotImplementedError


class CapabilityRequestIF:
    id: int
    state: str
    capability: CapabilityIF
    parameters: str
    future_products: str
    versions: List["CapabilityVersionIF"]

    def __json__(self, request: Any) -> dict:
        raise NotImplementedError


class CapabilityVersionIF:
    capability_request: CapabilityRequestIF
    version_number: int
    # eventually, the parameters should go here


class CapabilityExecutionIF:
    capability_version: CapabilityVersionIF
    state: str
    current_step: int

    def __json__(self, request: Any) -> dict:
        raise NotImplementedError


"""

Capability System / Services

"""


class CapabilityInfoIF(ABC):
    """
    Data access object that can look up and record information about capabilities
    and capability requests.
    """

    @abstractmethod
    def create_capability(
        self, name: CapabilityName, steps: List["CapabilityStepType"], max_jobs: int
    ) -> CapabilityIF:
        """
        Create new capability and save it in the database
        :param name: Name of new capability
        :param steps: List of capability steps required to execute the capability
        :param max_jobs: Max allowed number of concurrent executions of this kind of capability
        :return: Integer identifier for the capability
        """
        raise NotImplementedError

    @abstractmethod
    def create_capability_request(
        self,
        capability_name: str,
        parameters: List["ParameterIF"] = None,
        future_products: List[FutureProductIF] = None,
    ) -> CapabilityRequestIF:
        """
        Create new capability request and save it in the database
        :param capability_id: ID of the requested capability
        :param parameters: List
        :param future_products:
        :param versions:
        :return: Integer identifier of the request
        """
        raise NotImplementedError

    @abstractmethod
    def create_execution(self, request: CapabilityRequestIF) -> CapabilityExecutionIF:
        """
        Create new execution record for a request and save it in the database
        :param request: ID of the capability request
        :return: Integer identifier for the record
        """
        raise NotImplementedError

    @abstractmethod
    def lookup_capability(self, capability_name: str) -> CapabilityIF:
        raise NotImplementedError

    def lookup_execution(self, execution_id: int) -> CapabilityExecutionIF:
        raise NotImplementedError


class ProjectSettingsIF(ABC):
    """
    Class that deals with projects overriding capability settings and making multiple instances
    of a capability with the same name
    """

    pass


class CapabilityMatrixIF(ABC):
    """
    Maps CASA versions to version-specific templates, allowing us to support a variety of
    CASA versions
    """

    pass


class CapabilityQueueIF(ABC):
    """
    Organizes capability executions in a priority order and makes it possible to control
    the number of concurrent executions or pause execution altogether
    """

    @abstractmethod
    def enqueue(self, execution: CapabilityExecutionIF, priority: int):
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")

    @abstractmethod
    def dequeue(self):
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")

    @abstractmethod
    def pause(self):
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")


class CapabilityEngineIF(ABC):
    """
    Executes a prepare and run workflow step of a capability
    """

    @abstractmethod
    def execute(self):
        pass


class CapabilityServiceIF(ABC):
    """
    The capability service: clients access this to request capability runs
    """

    @abstractmethod
    def run_capability(self, request: CapabilityRequestIF):
        """
        Run a capability with the requested settings
        :param request: Capability request
        """
        raise NotImplementedError


class EstimationServiceIF(ABC):
    """
    Service that estimates how long a capability will take to complete
    """

    pass


class MetricsServiceIF(ABC):
    """
    Service that records metrics
    """

    pass


class ArchiveServiceIF(ABC):
    """
    Abstracts services that are needed from the archive system.
    """

    @abstractmethod
    def lookup_product(self, locator: ProductLocator) -> "ProductIF":
        """
        Look up a science product by its locator
        :param locator:  science product locator for this product
        :return:         science product
        """
        raise NotImplementedError(f"{self.__class__.__name__}.{inspect.stack()[0][3]}")


FieldName = str
FieldLabel = str


"""

Capability System / Helpers

"""


class CapabilitySequenceIF(ABC):
    pass


class EventTranslatorIF(ABC):
    """
    Translates a Workflow event into a Capability event
    """

    pass


class CapabilityStepIF(ABC):
    """
    A step in a capability sequence
    """

    pass


class FutureCapabilityResultIF(ProductIF):
    """
    Class that represents a product that will in the future result from a completed
    capability execution
    """

    pass


class ParameterIF(ABC):
    """
    Class that represents a set of capability parameters, a required input for the correct
    and complete execution of a capability
    """

    def json(self) -> str:
        """
        Convert to json-formatted string
        :return: JSON string
        """
        raise NotImplementedError
